@@ -5,8 +5,10 @@
 #include "primality.h"
 #include "prime_iterator.h"
 #include "utility.h"
-#include "small_factor.h"
+#include "pbrent63.h"
+#include "squfof126.h"
 #include "ecm.h"
+#include "tinyqs.h"
 #include "simpqs.h"
 
 #define _GMP_ECM_FACTOR(n, f, b1, ncurves) \
@@ -151,7 +153,8 @@ int factor(mpz_t input_n, mpz_t* pfactors[], int* pexponents[])
     while ( mpz_cmp_ui(n, tlim*tlim) > 0 && !_GMP_is_prob_prime(n) ) {
       int success = 0;
       int o = get_verbose_level();
-      UV nbits, B1 = 5000;
+      UV B1 = 5000;
+      UV nbits = mpz_sizeinbase(n, 2);
 
       /*
        * This set of operations is meant to provide good performance for
@@ -166,26 +169,40 @@ int factor(mpz_t input_n, mpz_t* pfactors[], int* pexponents[])
        * of many 12-digit or 14-digit primes should take under 10 seconds.
        */
 
-      if (mpz_cmp_ui(n, (unsigned long)(UV_MAX>>4)) < 0) {
-        UV ui_n = mpz_get_ui(n);
-        UV ui_factors[2];
-        if (!mpz_cmp_ui(n, ui_n)) {
-          success = racing_squfof_factor(ui_n, ui_factors, 200000)-1;
-          if (success) {
-            mpz_set_ui(f, ui_factors[0]);
-          } else {
-            if (o > 2) {gmp_printf("UV SQUFOF failed %Zd\n", n);}
-          }
-        }
-        if (success&&o) {gmp_printf("UV SQUFOF found factor %Zd\n", f);o=0;}
+      /* Handle small inputs here */
+      if (nbits <= 63) {
+        if (!success) success = pbrent63(n, f, 400000);
+        if (success&&o) {gmp_printf("UV Rho-Brent found factor %Zd\n", f);o=0;}
+      }
+      if (nbits >= 65 && nbits <= 126) {
+        if (!success) success = _GMP_pminus1_factor(n, f, 5000, 5000);
+        if (success&&o) {gmp_printf("p-1 (%dk) found factor %Zd\n",5000,f);o=0;}
+        if (!success) success = tinyqs(n, f);
+        if (success&&o) {gmp_printf("tinyqs found factor %Zd\n", f);o=0;}
+      }
+
+      /* It's possible the previous calls failed or weren't available */
+      if (nbits <= 53) {
+        if (!success)  success = squfof126(n, f, 400000);
+        if (success&&o) {gmp_printf("UV SQUFOF126 found factor %Zd\n", f);o=0;}
+      } else if (nbits <= 77) {
+        int sb1 = (nbits < 58) ?  1
+                : (nbits < 63) ?  2
+                : (nbits < 72) ?  4
+                               : 10;
+        if (!success)  success = _GMP_pminus1_factor(n, f, sb1*1000, sb1*10000);
+        if (success&&o) {gmp_printf("p-1 (%dk) found factor %Zd\n",sb1,f);o=0;}
+
+        if (!success)  success = squfof126(n, f, 1000000);
+        if (success&&o) {gmp_printf("SQUFOF126 found factor %Zd\n", f);o=0;}
       }
 
       /* Make sure it isn't a perfect power */
       if (!success)  success = (int)power_factor(n, f);
       if (success&&o) {gmp_printf("perfect power found factor %Zd\n", f);o=0;}
 
-      if (!success)  success = _GMP_pminus1_factor(n, f, 15000, 150000);
-      if (success&&o) {gmp_printf("p-1 (15k) found factor %Zd\n", f);o=0;}
+      if (!success)  success = _GMP_pminus1_factor(n, f, 20000, 200000);
+      if (success&&o) {gmp_printf("p-1 (20k) found factor %Zd\n", f);o=0;}
 
       /* Small ECM to find small factors */
       if (!success)  success = _GMP_ECM_FACTOR(n, f, 200, 4);
@@ -197,7 +214,6 @@ int factor(mpz_t input_n, mpz_t* pfactors[], int* pexponents[])
 
       /* Small p-1 */
       if (!success) {
-        nbits = mpz_sizeinbase(n, 2);
         if (nbits < 100 || nbits >= 160) {
           success = _GMP_pminus1_factor(n, f, 200000, 3000000);
           if (success&&o) {gmp_printf("p-1 (200k) found factor %Zd\n", f);o=0;}
@@ -276,9 +292,6 @@ int factor(mpz_t input_n, mpz_t* pfactors[], int* pexponents[])
       /*
       if (!success)  success = _GMP_pbrent_factor(n, f, 2, 512*1024*1024);
       if (success&&o) {gmp_printf("pbrent (2,512M) found factor %Zd\n", f);o=0;}
-
-      if (!success)  success = _GMP_squfof_factor(n, f, 256*1024*1024);
-      if (success&&o) {gmp_printf("squfof found factor %Zd\n", f);o=0;}
       */
 
       /* Our method of last resort: ECM with high bmax and many curves*/
@@ -436,7 +449,13 @@ int moebius(mpz_t n)
   int* exponents;
   int i, nfactors, result;
 
-  if (mpz_sgn(n) <= 0)       return 0;
+  if (mpz_sgn(n) < 0) {
+    mpz_neg(n,n);
+    result = moebius(n);
+    mpz_neg(n,n);
+    return result;
+  }
+  if (mpz_sgn(n) == 0) return 0;
   if (mpz_cmp_ui(n, 1) == 0) return 1;
 
   for (i = 0; i < 7; i++)
@@ -605,7 +624,7 @@ void znorder(mpz_t res, mpz_t a, mpz_t n)
 void znprimroot(mpz_t root, mpz_t n)
 {
   mpz_t t, phi, a, on, r, *factors;
-  int i, nfactors, *exponents;
+  int i, nfactors, *exponents, oddprime, k;
 
   mpz_set_ui(root, 0);
   if (mpz_cmp_ui(n, 4) <= 0) {
@@ -633,6 +652,9 @@ void znprimroot(mpz_t root, mpz_t n)
   mpz_divexact(on, on, r);
   mpz_mul(phi, phi, on);
 
+  mpz_sub_ui(r,n,1);
+  oddprime = (mpz_cmp(r,phi) == 0);
+
   mpz_clear(on);
   mpz_clear(r);
 
@@ -646,7 +668,8 @@ void znprimroot(mpz_t root, mpz_t n)
 
   for (mpz_set_ui(a,2);  mpz_cmp(a,n) < 0;  mpz_add_ui(a,a,1)) {
     if (!mpz_cmp_ui(a,4) || !mpz_cmp_ui(a,8) || !mpz_cmp_ui(a,9)) continue;
-    if (mpz_kronecker(a, n) == 0) continue;
+    k = mpz_kronecker(a, n);
+    if ( (oddprime && k != -1) || (!oddprime && k == 0) ) continue;
     for (i = 0; i < nfactors; i++) {
       mpz_powm(t, a, factors[i], n);
       if (mpz_cmp_ui(t, 1) == 0)
@@ -748,6 +771,62 @@ void ramanujan_tau(mpz_t res, mpz_t n)
   mpz_clear(t4); mpz_clear(t3); mpz_clear(t2); mpz_clear(t1); mpz_clear(t);
 }
 
+int is_semiprime(mpz_t n)
+{
+  int ret;
+  UV div, lim = 6000;
+  mpz_t t;
+
+  if (mpz_cmp_ui(n,6) < 0)
+    return (mpz_cmp_ui(n,4) == 0);
+
+  mpz_init(t);
+
+  div = _GMP_trial_factor(n, 2, lim);
+  if (div > 0) {
+    mpz_divexact_ui(t, n, div);
+    ret = _GMP_is_prime(t);
+    mpz_clear(t);
+    return !!ret;
+  }
+  /* No small divisors */
+  if (_GMP_BPSW(n))
+    { mpz_clear(t); return 0; }
+  if (mpz_ui_pow_ui(t,lim,3), mpz_cmp(n,t) < 0)
+    { mpz_clear(t); return 1; }
+  /* Number is composite, isn't tiny, and has no small divisors */
+  if (    0
+       || _GMP_pbrent_factor(n, t, 1, 15000)
+       || _GMP_pminus1_factor(n, t, 50000, 500000)
+       || _GMP_ECM_FACTOR(n, t,     800, 10)
+       || _GMP_ECM_FACTOR(n, t,    8000, 20)
+       || _GMP_ECM_FACTOR(n, t,   80000, 40)
+       || _GMP_ECM_FACTOR(n, t,  320000, 40)
+       || _GMP_ECM_FACTOR(n, t, 1000000, 80)
+     ) {
+    ret = _GMP_BPSW(t);
+    if (ret) {
+      mpz_divexact(t, n, t);
+      ret = _GMP_BPSW(t);
+    }
+    mpz_clear(t);
+    return !!ret;
+  }
+  /* No luck finding a small factor.  Do it the hard way. */
+  {
+    mpz_t* factors;
+    int* exponents;
+    int nfactors, i, j;
+
+    nfactors = factor(n, &factors, &exponents);
+    for (i = 0, j = 0; i < nfactors; i++)
+      j += exponents[i];
+    clear_factors(nfactors, &factors, &exponents);
+    mpz_clear(t);
+    return (j == 2);
+  }
+}
+
 
 /*****************************************************************************/
 
@@ -925,10 +1004,9 @@ int _GMP_prho_factor(mpz_t n, mpz_t f, UV a, UV rounds)
       mpz_mul(U, U, U);  mpz_add_ui(U, U, a);  mpz_tdiv_r(U, U, n);
       mpz_mul(V, V, V);  mpz_add_ui(V, V, a);  mpz_tdiv_r(V, V, n);
       mpz_mul(V, V, V);  mpz_add_ui(V, V, a);  mpz_tdiv_r(V, V, n);
-      if (mpz_cmp(U, V) >= 0)  mpz_sub(f, U, V);
-      else                     mpz_sub(f, V, U);
+      mpz_sub(f, U, V);
       mpz_mul(m, m, f);
-      mpz_tdiv_r(m, m, n);
+      if ((i%4) == ((inner-1)%4)) mpz_tdiv_r(m, m, n);
     }
     mpz_gcd(f, m, n);
     if (!mpz_cmp_ui(f, 1))
@@ -977,10 +1055,9 @@ int _GMP_pbrent_factor(mpz_t n, mpz_t f, UV a, UV rounds)
       mpz_set(saveXi, Xi);
       for (i = 0; i < dorounds; i++) {
         mpz_mul(t, Xi, Xi);  mpz_add_ui(t, t, a);  mpz_tdiv_r(Xi, t, n);
-        if (mpz_cmp(Xi, Xm) >= 0)  mpz_sub(f, Xi, Xm);
-        else                       mpz_sub(f, Xm, Xi);
-        mpz_mul(t, m, f);
-        mpz_tdiv_r(m, t, n);
+        mpz_sub(f, Xm, Xi);
+        mpz_mul(m, m, f);
+        if ((i%4) == ((dorounds-1)%4)) mpz_tdiv_r(m, m, n);
       }
       rleft -= dorounds;
       rounds -= dorounds;
@@ -998,20 +1075,30 @@ int _GMP_pbrent_factor(mpz_t n, mpz_t f, UV a, UV rounds)
       mpz_set(Xi, saveXi);
       do {
         mpz_mul(t, Xi, Xi);  mpz_add_ui(t, t, a);  mpz_tdiv_r(Xi, t, n);
-        if (mpz_cmp(Xi, Xm) >= 0)  mpz_sub(f, Xi, Xm);
-        else                       mpz_sub(f, Xm, Xi);
+        mpz_sub(f, Xm, Xi); if (mpz_sgn(f) < 0) mpz_add(f,f,n);
         mpz_gcd(f, f, n);
       } while (!mpz_cmp_ui(f, 1) && r-- != 0);
-      if ( (!mpz_cmp_ui(f, 1)) || (!mpz_cmp(f, n)) )  break;
     }
-    mpz_clear(Xi); mpz_clear(Xm); mpz_clear(m); mpz_clear(saveXi); mpz_clear(t);
-    return 1;
+    break;
   }
   mpz_clear(Xi); mpz_clear(Xm); mpz_clear(m); mpz_clear(saveXi); mpz_clear(t);
-  mpz_set(f, n);
-  return 0;
+  if (!mpz_cmp_ui(f, 1) || !mpz_cmp(f, n)) {
+    mpz_set(f, n);
+    return 0;
+  }
+  return 1;
 }
 
+/* References for P-1:
+ *  Montgomery 1987:  https://cr.yp.to/bib/1987/montgomery.pdf
+ *  Brent 1990:       http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.127.4316
+ *
+ * The main advantage of this over ECM is that it is *much* lower overhead,
+ * so very cheap to run with relatively small B1,B2 values.  A disadvantage
+ * is no continuation method, so subsequent calls with larger B1,B2 will
+ * repeat all the previous work.  ECM is much better for harder factorisations,
+ * so we typically want to try a little p-1 then move to ECM (or QS).
+ */
 int _GMP_pminus1_factor(mpz_t n, mpz_t f, UV B1, UV B2)
 {
   mpz_t a, savea, t;
@@ -1029,20 +1116,12 @@ int _GMP_pminus1_factor(mpz_t n, mpz_t f, UV B1, UV B2)
   if (_verbose>2) gmp_printf("# p-1 trying %Zd (B1=%"UVuf" B2=%"UVuf")\n", n, (unsigned long)B1, (unsigned long)B2);
 
   /* STAGE 1
-   * Montgomery 1987 p249-250 and Brent 1990 p5 both indicate we can calculate
-   * a^m mod n where m is the lcm of the integers to B1.  This can be done
-   * using either
-   *    m = calc_lcm(B), b = a^m mod n
-   * or
-   *    calculate_b_lcm(b, B1, a, n);
-   *
-   * The first means raising a to a huge power then doing the mod, which is
-   * inefficient and can be _very_ slow on some machines.  The latter does
-   * one powmod for each prime power, which works pretty well.  Yet another
-   * way to handle this is to loop over each prime p below B1, calculating
+   * See Montgomery 1987 p249-250 or Brent 1990 p5.  We can take E to be the
+   * lcm of integers to B1, then gcd(a^E-1,n) may be a factor of n.  While
+   * we could actually calculate the LCM, it is quite inefficient to do so.
+   * There are various ways to speed it up, but generally we prefer to do it
+   * the way Brent indicates, which is one powmod for each prime p below B1,
    * a = a^(p^e) mod n, where e is the largest e such that p^e <= B1.
-   * My experience with GMP is that this last method is faster with large B1,
-   * sometimes a lot faster.
    *
    * One thing that can speed things up quite a bit is not running the GCD
    * on every step.  However with small factors this means we can easily end
@@ -1380,201 +1459,6 @@ int _GMP_holf_factor(mpz_t n, mpz_t f, UV rounds)
   return 0;
 }
 
-
-/*----------------------------------------------------------------------
- * GMP version of Ben Buhrow's public domain 9/24/09 implementation.
- * It uses ideas and code from Jason Papadopoulos, Scott Contini, and
- * Tom St. Denis.  Also see the papers of Stephen McMath, Daniel Shanks,
- * and Jason Gower.  Gower and Wagstaff is particularly useful:
- *    http://homes.cerias.purdue.edu/~ssw/squfof.pdf
- *--------------------------------------------------------------------*/
-
-static int shanks_mult(mpz_t n, mpz_t f)
-{
-   /*
-    * use shanks SQUFOF to factor N.
-    *
-    * return 0 if no factor found, 1 if found with factor in f1.
-    *
-    * Input should have gone through trial division to 5.
-    */
-
-   int result = 0;
-   unsigned long j=0;
-   mpz_t b0, bn, imax, tmp, Q0, Qn, P, i, t1, t2, S, Ro, So, bbn;
-
-   if (mpz_cmp_ui(n, 3) <= 0)
-     return 0;
-
-   if (mpz_perfect_square_p(n)) {
-     mpz_sqrt(f, n);
-     return 1;
-   }
-
-   mpz_init(b0);
-   mpz_init(bn);
-   mpz_init(imax);
-   mpz_init(tmp);
-   mpz_init(Q0);
-   mpz_init(Qn);
-   mpz_init(P);
-   mpz_init(i);
-   mpz_init(t1);
-   mpz_init(t2);
-   mpz_init(S);
-   mpz_init(Ro);
-   mpz_init(So);
-   mpz_init(bbn);
-
-   mpz_sqrt(b0, n);
-   mpz_sqrt(tmp, b0);
-   mpz_mul_ui(imax, tmp, 3);
-
-   /* set up recurrence */
-   mpz_set_ui(Q0, 1);
-   mpz_set(P, b0);
-   mpz_mul(tmp, b0, b0);
-   mpz_sub(Qn, n, tmp);
-
-   mpz_add(tmp, b0, P);
-   mpz_tdiv_q(bn, tmp, Qn);
-
-   mpz_set_ui(i, 0);
-   while (1) {
-      j=0;
-      while (1) {
-         mpz_set(t1, P);   /* hold Pn for this iteration */
-         mpz_mul(tmp, bn, Qn);
-         mpz_sub(P, tmp, P);
-         mpz_set(t2, Qn);  /* hold Qn for this iteration */
-         mpz_sub(tmp, t1, P);
-         mpz_mul(tmp, tmp, bn);
-         mpz_add(Qn, Q0, tmp);
-         mpz_set(Q0, t2);  /* remember last Q */
-         mpz_add(tmp, b0, P);
-         mpz_tdiv_q(bn, tmp, Qn);
-
-         if (mpz_even_p(i)) {
-           if (mpz_perfect_square_p(Qn)) {
-             mpz_add_ui(i, i, 1);
-             break;
-           }
-         }
-         mpz_add_ui(i, i, 1);
-
-         if (mpz_cmp(i, imax) >= 0) {
-           result = 0;
-           goto end;
-         }
-      }
-
-      /* reduce to G0 */
-      mpz_sqrt(S, Qn);
-      mpz_sub(tmp, b0, P);
-      mpz_tdiv_q(tmp, tmp, S);
-      mpz_mul(tmp, S, tmp);
-      mpz_add(Ro, P, tmp);
-      mpz_mul(tmp, Ro, Ro);
-      mpz_sub(tmp, n, tmp);
-      mpz_tdiv_q(So, tmp, S);
-      mpz_add(tmp, b0, Ro);
-      mpz_tdiv_q(bbn, tmp, So);
-
-      /* search for symmetry point */
-      while (1) {
-         mpz_set(t1, Ro);  /* hold Ro for this iteration */
-         mpz_mul(tmp, bbn, So);
-         mpz_sub(Ro, tmp, Ro);
-         mpz_set(t2, So);  /* hold So for this iteration */
-         mpz_sub(tmp, t1, Ro);
-         mpz_mul(tmp, bbn, tmp);
-         mpz_add(So, S, tmp);
-         mpz_set(S, t2);   /* remember last S */
-         mpz_add(tmp, b0, Ro);
-         mpz_tdiv_q(bbn, tmp, So);
-
-         /* check for symmetry point */
-         if (mpz_cmp(Ro, t1) == 0)
-            break;
-
-         /* this gets stuck very rarely, but it does happen. */
-         if (++j > 1000000000)
-         {
-            result = -1;
-            goto end;
-         }
-      }
-
-      mpz_gcd(t1, Ro, n);
-      if (mpz_cmp_ui(t1, 1) > 0) {
-         mpz_set(f, t1);
-         /* gmp_printf("GMP SQUFOF found factor after %Zd/%lu rounds: %Zd\n", i, j, f); */
-         result = 1;
-         goto end;
-      }
-   }
-
-   end:
-   mpz_clear(b0);
-   mpz_clear(bn);
-   mpz_clear(imax);
-   mpz_clear(tmp);
-   mpz_clear(Q0);
-   mpz_clear(Qn);
-   mpz_clear(P);
-   mpz_clear(i);
-   mpz_clear(t1);
-   mpz_clear(t2);
-   mpz_clear(S);
-   mpz_clear(Ro);
-   mpz_clear(So);
-   mpz_clear(bbn);
-   return result;
-}
-
-int _GMP_squfof_factor(mpz_t n, mpz_t f, UV rounds)
-{
-   const UV multipliers[] = {
-      3*5*7*11, 3*5*7,  3*5*7*11*13, 3*5*7*13, 3*5*7*11*17, 3*5*11,
-      3*5*7*17, 3*5,    3*5*7*11*19, 3*5*11*13,3*5*7*19,    3*5*7*13*17,
-      3*5*13,   3*7*11, 3*7,         5*7*11,   3*7*13,      5*7,
-      3*5*17,   5*7*13, 3*5*19,      3*11,     3*7*17,      3,
-      3*11*13,  5*11,   3*7*19,      3*13,     5,           5*11*13,
-      5*7*19,   5*13,   7*11,        7,        3*17,        7*13,
-      11,       1 };
-   const size_t sz_mul = sizeof(multipliers)/sizeof(multipliers[0]);
-   size_t i;
-   int result;
-   mpz_t t;
-
-   TEST_FOR_2357(n, f);
-   mpz_init(t);
-   mpz_set_ui(f, 1);
-
-   for (i = 0; i < sz_mul; i++) {
-      UV mult = multipliers[i];
-      /* Only run when 64*m^3 < n */
-      mpz_set_ui(t, mult);
-      mpz_pow_ui(t, t, 3);
-      mpz_mul_ui(t, t, 64);
-      if (mpz_cmp(t, n) >= 0)
-        continue;
-      /* Run with this multiplier */
-      mpz_mul_ui(t, n, mult);
-      result = shanks_mult(t, f);
-      if (result == -1)
-        continue;
-      if ( (result == 1) && (mpz_cmp_ui(f, mult) != 0) ) {
-        unsigned long gcdf = mpz_gcd_ui(NULL, f, mult);
-        mpz_divexact_ui(f, f, gcdf);
-        if (mpz_cmp_ui(f, 1) > 0)
-          break;
-      }
-   }
-   mpz_clear(t);
-   return (mpz_cmp_ui(f, 1) > 0);
-}
-
 /* See if n is a perfect power */
 UV power_factor(mpz_t n, mpz_t f)
 {
@@ -1611,7 +1495,7 @@ mpz_t * divisor_list(int *num_divisors, mpz_t n)
   int nfactors, ndivisors, i, j, k, count, *exponents;
 
   nfactors = factor(n, &factors, &exponents);
-  ndivisors = exponents[0] + 1;
+  ndivisors = 1 + ((nfactors > 0) ? exponents[0] : 0);
   for (i = 1; i < nfactors; i++)
     ndivisors *= (exponents[i] + 1);
 
@@ -1630,6 +1514,7 @@ mpz_t * divisor_list(int *num_divisors, mpz_t n)
       }
     }
   }
+  mpz_clear(mult);
   clear_factors(nfactors, &factors, &exponents);
 
   qsort(divs, ndivisors, sizeof(mpz_t), numcmp);
